/***************************************************************************
 *            element.c
 *
 *  Tue Dec 21 15:40:34 2004
 *  Copyright  2004  Jean-Pierre Lozi
 *  jean-pierre@lozi.org - http://gtuxball.lozi.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>

/* We need to include some local headers. */
#include "error.h"
#include "element.h"
#include "globals.h"

/* Private functions : */

/* This function calls a block, the data parameter can be used to give
 * additional arguments needed by the API, such as where the element should
 * be drawn. */
void element_draw_block( Element el, void * data );

/* This function calls a ball, the data parameter can be used to give
 * additional arguments needed by the API, such as where the element should
 * be drawn. */
void element_draw_ball( Element el, void * data );

/* This function calls the pad, the data parameter can be used to give
 * additional arguments needed by the API, such as where the element should
 * be drawn. */
void element_draw_pad( Element el, void * data );

/* The element structure */
struct element  {
	/* The element type */
	int type;
	/* The element's coordinates */
	int x, y;
	/* The element's width and height */
	int width,height;
	/* The element's color */
	int color;
	/* The element's additional data */
	void * data;
	/* The element drawing function */
	void ( *draw )( struct element * el, void * data );
};

/* This arrays contains links to all the blocks' image files. */
const char * blocks_image_files[] = {	"pixmaps/red-block.png",
						 	 			"pixmaps/green-block.png",
						 				"pixmaps/blue-block.png",
						 				"pixmaps/cyan-block.png",
						 				"pixmaps/magenta-block.png",
						 				"pixmaps/yellow-block.png",
						 				"pixmaps/purple-block.png" };

/* Link to the pad's image file. */
const char * pad_image_files[] = {	"pixmaps/pad1_left.png",
									"pixmaps/pad2_left.png",
									"pixmaps/pad1_right.png",
									"pixmaps/pad2_right.png"};

/* Link to the ball's image file. */
const char * ball_image_file = "pixmaps/ball.png";
									
/*
 * This functions creates and returns an Element of the given type.
 */
Element create_element( int type )
{
	Element obj;
	
	/* We shall not forget that Element is a pointer to the element structure. */
	if( !( obj = ( Element )malloc( sizeof( struct element ) ) ) )
		error( MEMORY_ERROR );
	
	/* Now, initializations depend on the type. */
	switch( type ) {
		
		/* We do not need to initialize the x and y coordinates for a block,
		 * as they will be changed as soon as it is created. */
		case BLOCK:
    		obj->type = BLOCK;
    		obj->width = BLOCK_WIDTH;
    		obj->height = BLOCK_HEIGHT;
			obj->data = NULL;
			obj->draw = &element_draw_block;
    		break;
		
		/* The ball objects' initialization is pretty straightforward... */
		case BALL:
			obj->type = BALL;
    		obj->x = BALL_INITIAL_X;
    		obj->y = BALL_INITIAL_Y;
    		obj->width = BALL_DIAMETER;
    		obj->height = BALL_DIAMETER;
			obj->data = NULL;
			obj->draw = &element_draw_ball;
    		break;
		
		/* and so is the pad objects' initialization. */
		case PAD:
			obj->type = PAD;
    		obj->x = PAD_INITIAL_X;
    		obj->y = PAD_INITIAL_Y;
    		obj->width = PAD_WIDTH;
    		obj->height = PAD_HEIGHT;
			obj->data = NULL;
			obj->draw = &element_draw_pad;
			break;
	}
	
	/* Finally, we return the newly created object. */
	return obj;
}

/*
 * This function destroys the given element.
 */
void element_destroy( Element *obj )
{
	/* First, if the data pointer points anywhere, free it... */
	if( ( *obj )->data ) free( (( *obj )->data) );
	/* ...then, free the element... */
	free( *obj );
	/* ...and then update the pointer. */
	*obj = NULL;
}

/*
 * This function allows to modify the element's integer attributes.
 */
void element_set_attribute( Element obj, int attribute, int value )
{
	/* Depending on the requested attribute, the right value is modified. */ 
	switch( attribute )
	{
		case ELEMENT_X:
			obj->x = value;
    		break;

		case ELEMENT_Y:
			obj->y = value;
			break;
	
		case ELEMENT_WIDTH:
			obj->width = value;
			break;
		
		case ELEMENT_HEIGHT:
			obj->height = value;
			break;
		
		case ELEMENT_COLOR:
			obj->color = value;
			break;
  	}
}

/*
 * This function allows to modify the element's data.
 */
void element_set_data( Element obj, void * data )
{
	/* We just update the data field. */
	obj->data = data;
}

/*
 * This function allows to access the element's integer attributes.
 */
int element_get_attribute( Element obj, int attribute )
{
	/* Depending on the requested attribute, the right field is returned. */
	switch( attribute )
	{
		case ELEMENT_TYPE:
			return obj->type;
    		break;
		
		case ELEMENT_X:
			return obj->x;
    		break;

		case ELEMENT_Y:
			return obj->y;
			break;
		
		case ELEMENT_WIDTH:
			return obj->width;
			break;
		
		case ELEMENT_HEIGHT:
			return obj->height;
			break;
		
		case ELEMENT_COLOR:
			return obj->color;
    		break;
	}
}

/*
 * This function allows to access the element's data.
 */
void * element_get_data( Element obj )
{
	/* We just return the right field. */
	return obj->data;
}

/*
 * This function allows to access the element's drawing function.
 */
Element_draw_function element_get_draw_function( Element obj )
{
	/* We just return the right field. */
	return obj->draw;
}

/*
 * This function allows to quickly move an element.
 */
void element_move( Element obj, int x_offset, int y_offset )
{
	obj->x += x_offset;
	obj->y += y_offset;
}

/* This function draws a block, the data parameter can be used to give
 * additional arguments needed by the API, such as where the element should
 * be drawn. */
void element_draw_block( Element el, void * data )
{
	graphics_draw_block( el->color, el->x, el->y,
							 el->width, el->height, data );
}

/* This function draws a ball, the data parameter can be used to give
 * additional arguments needed by the API, such as where the element should
 * be drawn. */
void element_draw_ball( Element el, void * data )
{
	graphics_draw_ball( el->x, el->y, el->width, el->height, data );
}

/* This function draws a pad, the data parameter can be used to give
 * additional arguments needed by the API, such as where the element should
 * be drawn. */
void element_draw_pad( Element el, void * data )
{
	graphics_draw_pad( el->x, el->y, el->width, el->height, data );
}
