/***************************************************************************
 *            game.c
 *
 *  Tue Dec 21 18:27:33 2004
 *  Copyright  2004  Jean-Pierre Lozi
 *  jean-pierre@lozi.org - http://gtuxball.lozi.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>

#include "graphics.h"
#include "element.h"
#include "score.h"
#include "game.h"
#include "globals.h"
#include "error.h"

/* Private functions */

/* This function returns a nonzero value iff the point (x,y) is a free
 * position. If so, it also stores the blocks coordinates in the
 * hit_block_i and hit_block_j variables. */
int is_empty( int x, int y, int * hit_block_i, int * hit_block_j );

/* This function returns a nonzero value iff the point (x,y) belongs to the
 * given rectangle. */
int point_in_rect(	int x, int y, int rect_x, int rect_y,
					int rect_width, int rect_height );
/* This function returns the number of blocks in the given level. */
int game_get_blocks_num_for_level( int level );
/* This function loads given level. */
void game_load_level( int level );
/* This function updates the best scores by trying to add the current score,
 * and shows the high scores dialog box if the current score is a high score. */
void game_update_best_scores( void );
/* This function comares two scores' diff times, returns 0 if they are equal,
 * -1 if the first is lesser than the second and 1 otherwise. */
int compare_diff_times( void * first, void * second );
/* This function comares two scores' number of blocks, returns 0 if they are
 * equal, -1 if the first is lesser than the second and 1 otherwise. */
int compare_numbers_of_blocks( void * first, void * second );

/* Globals */
int mouse_old_x = PAD_INITIAL_X;

/*
 * This function is called when the user chooses "New game" from the game menu.
 */
void game_on_new_game ( void )
{
	Element el;
	int i,j;
	int * move;
	
	/* We allocate an array containing the ball move each time on_timer
	 * is called. */
	if( !( move	= (int *)(malloc( 2 * sizeof( int ) ) ) ) )
		error( MEMORY_ERROR );
	
	/* We fist create the pad and the ball. */
	if( global_pad ) element_destroy( &global_pad );
	global_pad = create_element( PAD );
	if( global_ball ) element_destroy( &global_ball );
	global_ball = create_element( BALL );
	
	/* Then, we set the ball's data to its initial value : the initial ball
	   coordinates' incrementation per "time event". */
	move[0] = 0;
	move[1] = global_start_speed;
	element_set_data( global_ball, (void *)move );
	
	/* For each line... */
	for( i = 0 ; i < NUMBER_OF_LINES ; i++ )
	{
		/* For each column... */
		for( j = 0 ; j < NUMBER_OF_COLUMNS ; j++ )
		{
			/* We initialize the cell. */
			global_blocks[i][j] = NULL;
		}
	}
	
	/* We initialize all the global variables */
	global_lives = NUM_LIVES;
	global_speed = global_start_speed;
	global_current_level = global_game_start_level = global_start_level;
	global_remaining_blocks = game_get_blocks_num_for_level( global_current_level );
	global_total_blocks = 0;
	global_start_time = time( NULL );
	global_freeze_game_time = time( NULL );
	global_diff_time = 0;
	game_load_level( global_current_level );
	
	/* We wait for the signal. */
	global_state = WAITING_FOR_SIGNAL;
	/* We enable the right item. */
	graphics_menu_item_set_active( END_ITEM, 1 );
	
	/* We finally invalidate all the drawing area to provide feedback to the user. */
	graphics_invalidate_all();
}

/*
 * This function is called when the user chooses "Pause game" from the game menu.
 */
void game_on_pause_game ( void )
{
	/* If the game is in the state mode... */
	if( global_state == PLAYING )
	{
		/* ...we pause the game. */
		global_state = PAUSE;
		global_freeze_game_time = time( NULL );
	}
	/* Otherwise... */
	else 
	{
		/* ...we start the game. */
		graphics_add_timeout( TIMEOUT, &game_on_timer );
		global_state = PLAYING;
		global_diff_time += difftime( time( NULL ), global_freeze_game_time);
	}
	
	/* Finally, we redraw the drawing area */
	graphics_invalidate_all();
}

/*
 * This function is called when the user chooses "Scores..." from the game menu.
 */
void game_on_scores ( void )
{
	/* We just show the high scores dialog. */
	graphics_show_high_scores();
}

/*
 * This function is called when the user chooses "End game" from the game menu.
 */
void game_on_end_game ( void )
{
	/* We set the game state to game over. */
	global_state = GAME_OVER;
	
	/* We disable the right items. */
	graphics_menu_item_set_active( PAUSE_ITEM, 0 );
	graphics_menu_item_set_active( END_ITEM, 0 );
	
	/* Then, we redraw the whole drawing area... */
	graphics_invalidate_all();
	/* ...and update the best scores. */
	game_update_best_scores();
}

/* 
 * This function is called when any of the "Level" menu item is activated.
 */
void game_on_level( int level )
{
	/* We just update the start level... */
	global_start_level = level;
	/* ...and the start speed. */
	global_start_speed = GAME_INITIAL_SPEED +
							( ( level > NUMBER_OF_LEVELS / 2 )? 1 : 0);
}

/*
 * This function is called when the user presses a key.
 */
void game_on_key_press_event( int key )
{
	int x;
	
	/* If we are waiting for a signal, this event _is_ the signal. */
	if( global_state == WAITING_FOR_SIGNAL )
	{
		/* We enable the right item. */
		graphics_menu_item_set_active( PAUSE_ITEM, 1 );
		
		/* We want the game_on_timer_function to be called every 20 milliseconds... */
		graphics_add_timeout( TIMEOUT, &game_on_timer );
		
		/* ...we update the global variables... */
		global_state = PLAYING;
		
		/* ...we update the global_diff_time variable... */
		global_diff_time += difftime( time( NULL ), global_freeze_game_time);

		/* ...and invalidate the drawing area. */
		graphics_invalidate_all();
		
		return;
	}
	/* If the user is not playing and we are not waiting for a signal,
         * then we have nothing to do.	*/
	else if( global_state != PLAYING ) return;
	
	/* We get the pad's x coordinate. */
	x = element_get_attribute( global_pad, ELEMENT_X );
	
	/* Depending on the key... */
	switch( key )
	{
		/* We either increment or decrement the pad's x coordinate. */
		case LEFT_ARROW:
			x -= STEP;
			if( x < 0 ) x = 0;
			global_direction = LEFT;
			break;
		
		case RIGHT_ARROW:
			x += STEP;
			if( x > BOARD_WIDTH - PAD_WIDTH ) x = BOARD_WIDTH - PAD_WIDTH;
			global_direction = RIGHT;
			break;
		
		default: return;
	}
	
	/* Move the penguins' feet.*/
	global_feet = !global_feet;
	
	/* Then we update the pad... */
	element_set_attribute( global_pad, ELEMENT_X, x);
	
	/* ...and invalidate the drawing area. */
        graphics_invalidate_rect( 0, BOARD_HEIGHT - PAD_HEIGHT - 1, BOARD_WIDTH,
                                                          BOARD_HEIGHT );
}

/*
 * This function is called when the mouse moves.
 */
void game_on_motion_notify_event( int x, int y )
{
	/* If the user is not playing, we have nothing to do. */
	if( global_state != PLAYING && global_state != WAITING_FOR_SIGNAL ) return;
	
	/* We force the pad's x coordinate to stay within the board. */
	if( x < 0 ) x = 0;
	else if( x > BOARD_WIDTH - PAD_WIDTH ) x = BOARD_WIDTH - PAD_WIDTH;
	
	/* Then, we update the pad's x coordinate. */ 
	element_set_attribute( global_pad, ELEMENT_X, x);
	
	/* We make the penguin look in the right direction... */
	if( mouse_old_x < x ) global_direction = RIGHT;
	else global_direction = LEFT;
	
	/* ...and we store the mouse x coordinate to be able to get the right
	 * direction next time. */
	mouse_old_x = x;
	
	/* Move the penguins' feet.*/
	global_feet = !global_feet;
	
	/* And invalidate the drawing area. */
	graphics_invalidate_rect( 0, BOARD_HEIGHT - PAD_HEIGHT - 1, BOARD_WIDTH,
							  BOARD_HEIGHT );

}

/* 
 * This function is called every TIMEOUT milliseconds. If it returns 1, it is
 * called STEP seconds later, if it returns zero, it is no more called.
 */
int game_on_timer ( void )
{
	/* First we get the ball coordinates... */
	int old_x = element_get_attribute( global_ball, ELEMENT_X );
	int old_y = element_get_attribute( global_ball, ELEMENT_Y );
	/* ...then get the current ball offset per "timer event"... */
	int x_offset = ((int *)(element_get_data( global_ball )))[0];
	int y_offset = ((int *)(element_get_data( global_ball )))[1];
	/* ...and preview the next ball's position. */
	int x = old_x + x_offset;
	int y = old_y + y_offset;
	int i, j;
	
	i = j = -1;
	
	/* If this function is called in a state in which the game must be stopped,
	 * we stop the timer by returning 0. */
	if( global_state == GAME_OVER || global_state == WAITING_FOR_SIGNAL
		|| global_state == PAUSE || global_state == WON ) return 0;
	
	/* If the ball is under the pad...  */
	if( old_y + BALL_DIAMETER > 
		element_get_attribute( global_pad, ELEMENT_Y ) + LOSE_MARGIN )
	{
		int * move;
		
		/* ...then the user loses a live. */
		global_lives--;
		
		/* If there are no more lives... */
		if( !global_lives )
		{
			/* We stop the game... */
			game_on_end_game ();
			/* ...and the timer. */
			return 0;
		}
		
		/* Otherwise, enable the pause item... */
		graphics_menu_item_set_active( PAUSE_ITEM, 0 );
		/* ...destroy the ball... */
		element_destroy( &global_ball );
		/* ...re-create it... */
		global_ball = create_element( BALL );
		/* ...center the pad... */
		element_set_attribute( global_pad, ELEMENT_X, PAD_INITIAL_X );
		/* ...update the status... */
		global_state = WAITING_FOR_SIGNAL;
		/* ...update the global_freeze_game_time variable... */
		global_freeze_game_time = time( NULL );
		/* ...recreate the array containing the ball directions... */
		if( !( move = ( int * )( malloc( 2 * sizeof( int ) ) ) ) )
			error( MEMORY_ERROR );
		/* ...set it in the new ball's data field... */
		element_set_data( global_ball, (void *)move );
		/* ...set the starting directions... */
		((int *)(element_get_data( global_ball )))[0] = 0;
		((int *)(element_get_data( global_ball )))[1] = global_speed;
		/* ...redraw the whole drawing area... */
		graphics_invalidate_all();
		
		/* ...and stop the timer. */
		return 0;
	}
	
	/* If the ball collides with the pad... */
	if( point_in_rect( 	x + BALL_DIAMETER - 1, y + BALL_DIAMETER / 2 + 1,	
						element_get_attribute( global_pad, ELEMENT_X ),
						element_get_attribute( global_pad,
											   ELEMENT_Y ) - HIT_MARGIN,
						element_get_attribute( global_pad, ELEMENT_WIDTH ),
						element_get_attribute( global_pad, ELEMENT_HEIGHT )))
	{
		int new_x_offset, new_y_offset;
		double a, alpha, theta;
		
		/* We want to get a new x and y offset with the right angle, and a
		 * radius of SPEED, so that the ball does not slow down or speed up.
		 * a is the ball distance from the left of the pad, alpha the pad
		 * width. */
		a = ( x + BALL_DIAMETER / 2 + 1) 
				- element_get_attribute( global_pad,ELEMENT_X );
		alpha = element_get_attribute( global_pad, ELEMENT_WIDTH );
		
		/* Theta is the needed angle. */
		theta  = 3.141592653589793 * ( 1 - a / alpha );
		
		/* new_x_offset and new_y_offset are the new offsets. */
		new_x_offset = (int)(global_speed * cos( theta ));
		new_y_offset = -(int)(global_speed * sin( theta ));
		
		/* We do not want horizontal moves! */
		if( new_y_offset == 0 ) new_y_offset++;
		
		/* Then we update the offsets in the ball structure... */
		((int *)(element_get_data( global_ball )))[0] = new_x_offset;
		((int *)(element_get_data( global_ball )))[1] = new_y_offset;
		
		/* ...and do not stop the timer. */
		return 1;
	}

	/* If the center point of the left side of the ball or the center
	 * point of the right side of the ball collides with something... */
	if( !is_empty( x, y + BALL_DIAMETER / 2 + 1, &i, &j ) 
	  || !is_empty( x + BALL_DIAMETER - 1, y + BALL_DIAMETER / 2 + 1, &i, &j ) )
	{
		/* ...invert the ball's x_offset. */
		((int *)(element_get_data( global_ball )))[0] = -x_offset;
	}
	/* else if the center point of the top side of the ball or the center
	 * point of the bottom side of the ball collides with something... */
	else if( !is_empty( x + BALL_DIAMETER / 2 + 1, y, &i, &j )
		  || !is_empty( x + BALL_DIAMETER / 2 + 1, y + BALL_DIAMETER - 1, &i, &j ) )
	{
		/* ...invert the ball's y_offset. */
		((int *)(element_get_data( global_ball )))[1] = -y_offset;
	}
	/* Otherwise... */
	else
	{
		/* ...all gets one step further. */
		element_move( global_ball, x_offset, y_offset );
		/* Then we invalidate the drawing area... */
		
		graphics_invalidate_rect((( old_x < x ) ? old_x : x),
								 (( old_y < y ) ? old_y : y),
								 (( old_x < x ) ? x : old_x ) + BALL_DIAMETER,
								 (( old_y < y ) ? y : old_y ) + BALL_DIAMETER );
		
		/* ...and do not stop the timer. */
		return 1;
	}
	
	/* If the i and j values has been set... */
	if( i >= 0 )
	{
		/* ...then the block at this position was hit and must be destroyed. */
		element_destroy( &global_blocks[i][j] );
		global_blocks[i][j] = NULL;
		/* We decrease the number of remaining blocks. */
		global_remaining_blocks--;
		/* If no block remains... */
		if( global_remaining_blocks == 0 )
		{
			/* If the level is the last one... */
			if( global_current_level == NUMBER_OF_LEVELS - 1 )
			{
				/* The user has won the game. */
				global_state = WON;
				/* We disable the right items. */
				graphics_menu_item_set_active( PAUSE_ITEM, 0 );
				graphics_menu_item_set_active( END_ITEM, 0 );
				/* Then we redraw the area (to display the "You Win!"
				 * message)... */
				graphics_invalidate_all();
				/* ...we update the best scores... */
				game_update_best_scores();
				/* ...and do not stop the timer. */
				return 0;
			}
			/* Otherwise we increase the game speed and go to the next level.
			 * But we only increase the game speed every 3 levels. */
			if( global_current_level == NUMBER_OF_LEVELS / 2 ) global_speed++;
			
			game_load_level( ++global_current_level );
		}
		/* We just hit one more block... */
		global_total_blocks++;
		/* We finally invalidate the graphic area. */
		graphics_invalidate_all();
	}
	
	/* We do not stop the timer... */
	return 1;
}

/*
 * This function returns a nonzero value iff the point (x,y) is a free
 * position. If so, it also stores the blocks coordinates in the
 * hit_block_i and hit_block_j variables.
 */
int is_empty( int x, int y, int * hit_block_i, int * hit_block_j )
{
	int i, j;
	
	/* If we are off the game board, the point is not a free position. */
	if( x <= 0 || y <= 0 || x >= BOARD_WIDTH || y >= BOARD_HEIGHT )
		return 0;
	
	/* Otherwise, we locate the block where the point could be. */
	j = ( x - BLOCK_MARGIN ) / ( BLOCK_WIDTH + BLOCK_MARGIN );
	i = ( y - BLOCK_MARGIN ) / ( BLOCK_HEIGHT + BLOCK_MARGIN );
	/* We check that the point is actually in the block. */
	if( (i >= 0) && (i < NUMBER_OF_LINES) && (j >= 0)
		&& (j < NUMBER_OF_COLUMNS) && global_blocks[i][j]
		&& point_in_rect( x, y,
			element_get_attribute( global_blocks[i][j], ELEMENT_X ),
			element_get_attribute( global_blocks[i][j], ELEMENT_Y ),
			element_get_attribute( global_blocks[i][j], ELEMENT_WIDTH ),
			element_get_attribute( global_blocks[i][j], ELEMENT_HEIGHT )))
			{
				/* If it is, then we store the block's position... */
				*hit_block_i = i;
				*hit_block_j = j;
				/* ...and stop searching. */
				return 0;
			}
	
	/* Otherwise, the point is free. */
	return 1;
}

/*
 * This function returns a nonzero value iff the point (x,y) belongs to the
 * given rectangle.
 */
int point_in_rect(	int x, int y, int rect_x, int rect_y,
					int rect_width, int rect_height )
{
	/* This is pretty straightforward... */
	return ( x >= rect_x	&& y >= rect_y
							&& x <= ( rect_x + rect_width ) 
							&& y <= ( rect_y + rect_height ) );
}

/*
 * This function returns the number of blocks in the given level.
 */
int game_get_blocks_num_for_level( int level )
{
	int i, j, k = 0;
	
	/* For each line... */
	for( i = 0 ; i < NUMBER_OF_LINES ; i++ )
		/* For each column... */
		for( j = 0 ; j < NUMBER_OF_COLUMNS ; j++ )
			/* Count! */
			if( global_levels[level][i][j] ) k++;
	/* Finally, return the number of blocks in the given level. */
	return k;
}

/*
 * This function loads given level.
 */
void game_load_level( int level )
{
	int i, j;
	Element el;
	
	/* For each line... */
	for( i = 0 ; i < NUMBER_OF_LINES ; i++ )
	{
		/* For each column... */
		for( j = 0 ; j < NUMBER_OF_COLUMNS ; j++ )
		{
			/* If there is no block, therefore leave this cell of global_blocks empty. */
			if( global_levels[level][i][j] == 0x00 )
				global_blocks[i][j] = NULL;
			/* Otherwise, fill it: */
			else
			{
				if( global_blocks[i][j] ) element_destroy( &global_blocks[i][j] );
				/* We first create the element... */
				el = create_element( BLOCK );
				/* ...then set its color... */
				element_set_attribute( el,  ELEMENT_COLOR,
											global_levels[level][i][j]);
				/* ...set its location... */
				element_set_attribute( el,	ELEMENT_X,
											j * (BLOCK_WIDTH + BLOCK_MARGIN)
											+ BLOCK_MARGIN );
				element_set_attribute( el, ELEMENT_Y, i * (BLOCK_HEIGHT
														+ BLOCK_MARGIN));
				/* ...and store it in the right cell. */
				global_blocks[i][j] = el;
			}
		}
	}
	/* The number of remaining blocks in this level is the total number of 
	 * blocks in the level, as we just loaded it. */
	global_remaining_blocks = game_get_blocks_num_for_level( level );
}
	
/*
 * This function updates the best scores by trying to add the current score,
 * and shows the high scores dialog box if the current score is a high score.
 */
void game_update_best_scores( void )
{
	FILE *file;
	Score * scores;
	int i = 0, j = 0;
	int rank;
	char name[64];
	
	/* First, we read the high scores file. */
	scores = score_read_scores ();
	/* Then, we realloc the scores array to provide the space needed to insert
	 * a new score. */
	scores = ( Score * )realloc( scores, (sizeof( Score ) * 11 ) );
	/* We create the new high score, and insert it into the array. */
	scores[NUMBER_OF_SCORES] = score_create( "TempUser",
											 global_total_blocks, 
											 global_current_level,
											 global_game_start_level,
							   				 difftime(  time(NULL),
											 global_start_time ) - global_diff_time,
											 global_start_time );
	
	/* We sort the array by numbers of blocks first, and then by diff_times
	 * (which logically makes us call the two sorts in the reverse order. */ 
	qsort( scores, NUMBER_OF_SCORES + 1, sizeof( Score ), compare_diff_times );
	qsort( scores, NUMBER_OF_SCORES + 1, sizeof( Score ), compare_numbers_of_blocks );
	
	/* Through each score... */
	for( i = 0 ; i < NUMBER_OF_SCORES + 1 ; i++ )
	{
		/* We look for the current score. */
		char * str = ( char * )score_get_attribute( scores[i], SCORE_NAME );
		/* If we found it, we store its position. */
		if( !strcmp( str, "TempUser" )) rank = i;
	}
	/* If the current score is the last one, then we just leave. */
	if( rank == NUMBER_OF_SCORES ) return;
	
	/* Otherwise, we request the user name. */
	graphics_ask_name( name );
	
	/* We don't want the user to use our internal name! Anyway we just update
	 * the name in our score, either the given name, or the default user name
	 * if the provided name is our internal name. */
	if( !strcmp( name, "TempUser" ) || ( strlen( name ) < 1 ) || ( strrchr( name, ':' ) != NULL ) )
		strcpy( ( char * )score_get_attribute( scores[rank], SCORE_NAME ),
				DEFAULT_USER_NAME);
	else
		strcpy( ( char * )score_get_attribute( scores[rank], SCORE_NAME ),
				name);
	
	
	/* Then, we remove the high scores file to erase its contents, and
	 * recreate it. */
	remove( HIGH_SCORES_FILE );
	file = fopen( HIGH_SCORES_FILE, "w");
	
	/* If we can't open the file, we call the error function. */
	if( file == NULL ) error( HIGH_SCORES_OPEN_FILE_ERROR );
	/* Otherwise, we fill the file in. */
	for( i = 0; i < NUMBER_OF_SCORES + 1 ; i++ )
	{
		fprintf( file, "%s:%d:%d:%d:%d:%d:",
				 ( char * )score_get_attribute( scores[i], SCORE_NAME ),
				 ( int )score_get_attribute( scores[i], SCORE_NUMBER_OF_BLOCKS ),
				 ( int )score_get_attribute( scores[i], SCORE_START_LEVEL ),
				 ( int )score_get_attribute( scores[i], SCORE_LEVEL ),
				 ( long )score_get_attribute( scores[i], SCORE_DIFF_TIME ),
				 ( long )score_get_attribute( scores[i], SCORE_START_TIME ));	
	}
	/* Then we close the file. */
	fclose( file );
	
	/* Finally, we show the high scores window... */
	graphics_show_high_scores();
	
	/* ...and free the allocated memory. */
	for( i = NUMBER_OF_SCORES; i >= 0 ; i-- ) free( scores[i] );
	free( scores );	
}
	
/* 
 * This function comares two scores' diff times, returns 0 if they are equal,
 * -1 if the first is lesser than the second and 1 otherwise.
 */
int compare_diff_times( void * first, void * second )
{
	/* We first get the first diff_time... */
	int first_diff_time = ( int )score_get_attribute( (*((Score*)first)), 
													   SCORE_DIFF_TIME );
	/* ...the second one... */
	int second_diff_time = ( int )score_get_attribute( (*((Score*)second)),
														SCORE_DIFF_TIME );
	/* ...then compare them, and return the right result */
	if( first_diff_time == second_diff_time ) return 0;
	else if( first_diff_time < second_diff_time ) return -1;
	else return 1;
}

/* 
 * This function comares two scores' number of blocks, returns 0 if they are
 * equal, -1 if the first is lesser than the second and 1 otherwise.
 */
int compare_numbers_of_blocks( void * first, void * second )
{
	/* We first get the first diff_time... */
	int first_numbers_of_blocks = 
	  ( int )score_get_attribute( (*((Score*)first)), SCORE_NUMBER_OF_BLOCKS );
	/* ...the second one... */
	int second_numbers_of_blocks =
	  ( int )score_get_attribute( (*((Score*)second)), SCORE_NUMBER_OF_BLOCKS );
	/* ...then compare them, and return the right result */
	if( first_numbers_of_blocks == second_numbers_of_blocks ) return 0;
	else if( first_numbers_of_blocks < second_numbers_of_blocks ) return 1;
	else return -1;
}
